@@ -4,6 +4,16 @@ The Benchmark module with a unified data interface to evaluate graph matching me
 If you are interested in the performance and the deep learning framework, please refer to our `ThinkMatch project <https://github.com/Thinklab-SJTU/ThinkMatch>`_.
 """
 
+# Copyright (c) 2022 Thinklab@SJTU
+# pygmtools is licensed under Mulan PSL v2.
+# You can use this software according to the terms and conditions of the Mulan PSL v2.
+# You may obtain a copy of Mulan PSL v2 at:
+# http://license.coscl.org.cn/MulanPSL2
+# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
+# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
+# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
+# See the Mulan PSL v2 for more details.
+
 import tempfile
 import shutil
 import itertools
@@ -23,6 +33,11 @@ class Benchmark:
        ``'inclusion'`` is only for 2GM and refers to filtering only one graph to make its nodes a subset of the other graph,
        and ``'unfiltered'`` refers to retaining all nodes in all graphs
     :param args: keyword settings for specific dataset
+
+    .. note::
+            Ground truth cache is saved only when the parameter ``sets`` is ``'test'``, so
+            the functions ``eval()`` and ``eval_cls()`` are only for ``'test'`` set.
+
     """
 
     def __init__(self, name, sets, obj_resize=(256, 256), problem='2GM', filter='intersection', **args):
@@ -355,14 +370,24 @@ class Benchmark:
 
         return num_list
 
-    def eval(self, prediction, classes, verbose=False):
+    def eval(self, prediction, classes, verbose=False, rm_gt_cache=True):
         r"""
         Evaluate test results and compute matching accuracy and coverage.
 
         :param prediction: list, prediction result, like ``[{'ids': (id1, id2), 'cls': cls, 'permmat': np.array or scipy.sparse}, ...]``
         :param classes: list of evaluated classes
         :param verbose: bool, whether to print the result
+        :param rm_gt_cache: bool, whether to remove ground truth cache
         :return: evaluation result in each class and their averages, including p, r, f1 and their standard deviation and coverage
+
+        .. note::
+            If there are duplicate data pair in ``prediction``, this function will only evaluate the first pair and
+            expect that this pair is also the first fetched pair. Therefore, it is recommended that ``prediction`` is
+            built in an ordered manner, and not shuffled.
+
+        .. note::
+            Ground truth cache is saved when data pairs are fetched, and should be removed after evaluation. Make sure
+            all data pairs are evaluated at once, i.e., ``prediction`` should contain all fetched data pairs.
         """
 
         with open(self.data_list_path) as f1:
@@ -462,6 +487,8 @@ class Benchmark:
                                                         result['mean']['recall'], result['mean']['recall_std'],
                                                         result['mean']['f1'], result['mean']['f1_std']
                                                         )))
+        if rm_gt_cache:
+            self.rm_gt_cache(last_epoch=False)
         return result
 
     def eval_cls(self, prediction, cls, verbose=False):
@@ -472,6 +499,15 @@ class Benchmark:
         :param cls: str, evaluated class
         :param verbose: bool, whether to print the result
         :return: evaluation result on the specified class, including p, r, f1 and their standard deviation and coverage
+
+        .. note::
+            If there are duplicate data pair in ``prediction``, this function will only evaluate the first pair and
+            expect that this pair is also the first fetched pair. Therefore, it is recommended that ``prediction`` is
+            built in an ordered manner, and not shuffled. Same as the function ``eval``.
+
+        .. note::
+            This function will not automatically remove ground truth cache. However, you can still mannually call the
+            class function ``rm_gt_cache`` to remove groud truth cache after evaluation.
         """
 
         with open(self.data_list_path) as f1:
@@ -537,9 +573,9 @@ class Benchmark:
 
     def rm_gt_cache(self, last_epoch=False):
         r"""
-        Remove ground truth cache.
+        Remove ground truth cache. It is recommended to call this function after evaluation.
 
-        :param last_epoch: Boolean variable, whether this epoch is last epoch; if true, the directory of cache will also be removed.
+        :param last_epoch: bool, whether this epoch is last epoch; if true, the directory of cache will also be removed, and no more data should be evaluated
         """
         if os.path.exists(self.gt_cache_path):
             shutil.rmtree(self.gt_cache_path)
