@@ -2,7 +2,7 @@ r"""
 Classic (learning-free) **two-graph matching** solvers. These two-graph matching solvers are recommended to solve
 matching problems with two explicit graphs, or problems formulated as Quadratic Assignment Problem (QAP).
 
-The two-graph matching problem considers both nodes and edges, formulated as a QAP:
+The two-graph matching problem considers both nodes and edges, formulated as a QAP (Lawler's):
 
 .. math::
 
@@ -10,6 +10,16 @@ The two-graph matching problem considers both nodes and edges, formulated as a Q
     s.t. \quad &\mathbf{X} \in \{0, 1\}^{n_1\times n_2}, \ \mathbf{X}\mathbf{1} = \mathbf{1}, \ \mathbf{X}^\top\mathbf{1} \leq \mathbf{1}
 """
 
+# Copyright (c) 2022 Thinklab@SJTU
+# pygmtools is licensed under Mulan PSL v2.
+# You can use this software according to the terms and conditions of the Mulan PSL v2.
+# You may obtain a copy of Mulan PSL v2 at:
+# http://license.coscl.org.cn/MulanPSL2
+# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
+# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
+# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
+# See the Mulan PSL v2 for more details.
+
 import importlib
 import pygmtools
 from pygmtools.utils import NOT_IMPLEMENTED_MSG, _check_shape, _get_shape, _unsqueeze, _squeeze, _check_data_type
@@ -19,7 +29,7 @@ def sm(K, n1=None, n2=None, n1max=None, n2max=None, x0=None,
        max_iter: int=50,
        backend=None):
     r"""
-    Spectral Graph Matching solver for graph matching (QAP).
+    Spectral Graph Matching solver for graph matching (Lawler's QAP).
     This algorithm is also known as Power Iteration method, because it works by computing the leading
     eigenvector of the input affinity matrix by power iteration.
 
@@ -209,6 +219,69 @@ def sm(K, n1=None, n2=None, n1max=None, n2max=None, x0=None,
             # Accuracy
             >>> (pygm.hungarian(X) * X_gt).sum() / X_gt.sum()
             jt.Var([1.], dtype=float32)
+            
+            # This solver supports gradient back-propogation
+            >>> from jittor import nn
+            >>> class Model(nn.Module):
+            ...     def __init__(self, K):
+            ...         self.K = K
+            ...     def execute(self, K, n1, n2):
+            ...         X = pygm.sm(K, n1, n2)
+            ...         return X
+
+            >>> model = Model(K)
+            >>> optim = nn.SGD(model.parameters(), lr=0.1)
+            >>> X = model(K, n1, n2)
+            >>> loss = X.sum()
+            >>> optim.step(loss)
+            >>> len(jt.nonzero(K.opt_grad(optim)))
+            2560
+
+    .. dropdown:: Tensorflow Example
+
+        ::
+
+            >>> import tensorflow as tf
+            >>> import pygmtools as pygm
+            >>> pygm.BACKEND = 'tensorflow'
+            >>> _ = tf.random.set_seed(1)
+
+            # Generate a batch of isomorphic graphs
+            >>> batch_size = 10
+            >>> X_gt = tf.Variable(tf.zeros([batch_size, 4, 4]))
+            >>> indices = tf.stack([tf.range(4),tf.random.shuffle(tf.range(4))], axis=1)
+            >>> updates = tf.ones([4])
+            >>> for i in range(batch_size):
+            ...     _ = X_gt[i].assign(tf.tensor_scatter_nd_update(X_gt[i], indices, updates))
+            >>> A1 = tf.random.uniform([batch_size, 4, 4])
+            >>> A2 = tf.matmul(tf.matmul(tf.transpose(X_gt, perm=[0, 2, 1]), A1), X_gt)
+            >>> n1 = n2 = tf.constant([4] * batch_size)
+
+            # Build affinity matrix
+            >>> conn1, edge1, ne1 = pygm.utils.dense_to_sparse(A1)
+            >>> conn2, edge2, ne2 = pygm.utils.dense_to_sparse(A2)
+            >>> import functools
+            >>> gaussian_aff = functools.partial(pygm.utils.gaussian_aff_fn, sigma=1.) # set affinity function
+            >>> K = pygm.utils.build_aff_mat(None, edge1, conn1, None, edge2, conn2, n1, None, n2, None, edge_aff_fn=gaussian_aff)
+
+            # Solve by SM. Note that X is normalized with a squared sum of 1
+            >>> X = pygm.sm(K, n1, n2)
+            >>> tf.reduce_sum((X ** 2), axis=[1, 2])
+            <tf.Tensor: shape=(10,), dtype=float32, numpy=
+            array([1.        , 1.0000001 , 1.        , 0.9999999 , 1.        ,
+                   1.        , 1.0000001 , 0.99999994, 1.        , 0.9999998 ],
+                  dtype=float32)>
+
+            # Accuracy
+            >>> tf.reduce_sum((pygm.hungarian(X) * X_gt))/ tf.reduce_sum(X_gt)
+            <tf.Tensor: shape=(), dtype=float32, numpy=1.0>
+
+            This solver supports gradient back-propogation
+            >>> K = tf.Variable(K)
+            >>> with tf.GradientTape() as tape:
+            ...     y = tf.reduce_sum(pygm.sm(K, n1, n2))
+            ...     len(tf.where(tape.gradient(y, K)))
+            2560
 
     .. note::
         If you find this graph matching solver useful for your research, please cite:
@@ -246,7 +319,7 @@ def sm(K, n1=None, n2=None, n1max=None, n2max=None, x0=None,
     try:
         mod = importlib.import_module(f'pygmtools.{backend}_backend')
         fn = mod.sm
-    except ModuleNotFoundError and AttributeError:
+    except (ModuleNotFoundError, AttributeError):
         raise NotImplementedError(
             NOT_IMPLEMENTED_MSG.format(backend)
         )
@@ -261,8 +334,8 @@ def rrwm(K, n1=None, n2=None, n1max=None, n2max=None, x0=None,
          max_iter: int=50, sk_iter: int=20, alpha: float=0.2, beta: float=30,
          backend=None):
     r"""
-    Reweighted Random Walk Matching (RRWM) solver for graph matching (QAP). This algorithm is implemented by power
-    iteration with Sinkhorn reweighted jumps.
+    Reweighted Random Walk Matching (RRWM) solver for graph matching (Lawler's QAP). This algorithm is implemented by
+    power iteration with Sinkhorn reweighted jumps.
 
     The official matlab implementation is available at https://cv.snu.ac.kr/research/~RRWM/
 
@@ -455,6 +528,69 @@ def rrwm(K, n1=None, n2=None, n1max=None, n2max=None, x0=None,
             # Accuracy
             >>> (pygm.hungarian(X) * X_gt).sum() / X_gt.sum()
             jt.Var([1.], dtype=float32)
+            
+            # This solver supports gradient back-propogation
+            >>> from jittor import nn
+            >>> class Model(nn.Module):
+            ...     def __init__(self, K):
+            ...         self.K = K
+            ...     def execute(self, K, n1, n2, beta):
+            ...         X = pygm.rrwm(K, n1, n2, beta=beta)
+            ...         return X
+
+            >>> model = Model(K)
+            >>> optim = nn.SGD(model.parameters(), lr=0.1)
+            >>> X = model(K, n1, n2, beta=100)
+            >>> loss = X.sum()
+            >>> optim.step(loss)
+            >>> len(jt.nonzero(K.opt_grad(optim)))
+            1536
+
+    .. dropdown:: Tensorflow Example
+
+        ::
+
+            >>> import tensorflow as tf
+            >>> import pygmtools as pygm
+            >>> pygm.BACKEND = 'tensorflow'
+            >>> _ = tf.random.set_seed(1)
+
+            # Generate a batch of isomorphic graphs
+            >>> batch_size = 10
+            >>> X_gt = tf.Variable(tf.zeros([batch_size, 4, 4]))
+            >>> indices = tf.stack([tf.range(4),tf.random.shuffle(tf.range(4))], axis=1)
+            >>> updates = tf.ones([4])
+            >>> for i in range(batch_size):
+            ...     _ = X_gt[i].assign(tf.tensor_scatter_nd_update(X_gt[i], indices, updates))
+            >>> A1 = tf.random.uniform([batch_size, 4, 4])
+            >>> A2 = tf.matmul(tf.matmul(tf.transpose(X_gt, perm=[0, 2, 1]), A1), X_gt)
+            >>> n1 = n2 = tf.constant([4] * batch_size)
+
+            # Build affinity matrix
+            >>> conn1, edge1, ne1 = pygm.utils.dense_to_sparse(A1)
+            >>> conn2, edge2, ne2 = pygm.utils.dense_to_sparse(A2)
+            >>> import functools
+            >>> gaussian_aff = functools.partial(pygm.utils.gaussian_aff_fn, sigma=1.) # set affinity function
+            >>> K = pygm.utils.build_aff_mat(None, edge1, conn1, None, edge2, conn2, n1, None, n2, None, edge_aff_fn=gaussian_aff)
+
+            # Solve by RRWM. Note that X is normalized with a sum of 1
+            >>> X = pygm.rrwm(K, n1, n2, beta=100)
+            >>> tf.reduce_sum(X, axis=[1, 2])
+            <tf.Tensor: shape=(10,), dtype=float32, numpy=
+            array([1.        , 1.        , 1.        , 0.99999994, 1.        ,
+                   1.        , 1.        , 0.99999994, 0.99999994, 1.0000001 ],
+                  dtype=float32)>
+
+            # Accuracy
+            >>> tf.reduce_sum((pygm.hungarian(X) * X_gt)) / tf.reduce_sum(X_gt)
+            <tf.Tensor: shape=(), dtype=float32, numpy=1.0>
+
+            # This solver supports gradient back-propogation
+            >>> K = tf.Variable(K)
+            >>> with tf.GradientTape(persistent=True) as tape:
+            ...     y = tf.reduce_sum(pygm.rrwm(K, n1, n2, beta=100))
+            ...     len(tf.where(tape.gradient(y, K)))
+            768
 
     .. note::
         If you find this graph matching solver useful in your research, please cite:
@@ -495,7 +631,7 @@ def rrwm(K, n1=None, n2=None, n1max=None, n2max=None, x0=None,
     try:
         mod = importlib.import_module(f'pygmtools.{backend}_backend')
         fn = mod.rrwm
-    except ModuleNotFoundError and AttributeError:
+    except (ModuleNotFoundError, AttributeError):
         raise NotImplementedError(
             NOT_IMPLEMENTED_MSG.format(backend)
         )
@@ -510,7 +646,7 @@ def ipfp(K, n1=None, n2=None, n1max=None, n2max=None, x0=None,
          max_iter: int=50,
          backend=None):
     r"""
-    Integer Projected Fixed Point (IPFP) method for graph matching (QAP).
+    Integer Projected Fixed Point (IPFP) method for graph matching (Lawler's QAP).
 
     :param K: :math:`(b\times n_1n_2 \times n_1n_2)` the input affinity matrix, :math:`b`: batch size.
     :param n1: :math:`(b)` number of nodes in graph1 (optional if n1max is given, and all n1=n1max).
@@ -687,6 +823,46 @@ def ipfp(K, n1=None, n2=None, n1max=None, n2max=None, x0=None,
             >>> (pygm.hungarian(X) * X_gt).sum() / X_gt.sum()
             jt.Var([1.], dtype=float32)
 
+    .. dropdown:: Tensorflow Example
+
+        ::
+
+            >>> import tensorflow as tf
+            >>> import pygmtools as pygm
+            >>> pygm.BACKEND = 'tensorflow'
+            >>> _ = tf.random.set_seed(1)
+
+            # Generate a batch of isomorphic graphs
+            >>> batch_size = 10
+            >>> X_gt = tf.Variable(tf.zeros([batch_size, 4, 4]))
+            >>> indices = tf.stack([tf.range(4),tf.random.shuffle(tf.range(4))], axis=1)
+            >>> updates = tf.ones([4])
+            >>> for i in range(batch_size):
+            ...     _ = X_gt[i].assign(tf.tensor_scatter_nd_update(X_gt[i], indices, updates))
+            >>> A1 = tf.random.uniform([batch_size, 4, 4])
+            >>> A2 = tf.matmul(tf.matmul(tf.transpose(X_gt, perm=[0, 2, 1]), A1), X_gt)
+            >>> n1 = n2 = tf.constant([4] * batch_size)
+
+            # Build affinity matrix
+            >>> conn1, edge1, ne1 = pygm.utils.dense_to_sparse(A1)
+            >>> conn2, edge2, ne2 = pygm.utils.dense_to_sparse(A2)
+            >>> import functools
+            >>> gaussian_aff = functools.partial(pygm.utils.gaussian_aff_fn, sigma=1.) # set affinity function
+            >>> K = pygm.utils.build_aff_mat(None, edge1, conn1, None, edge2, conn2, n1, None, n2, None, edge_aff_fn=gaussian_aff)
+
+            # Solve by IPFP
+            >>> X = pygm.ipfp(K, n1, n2)
+            >>> X[0]
+            <tf.Tensor: shape=(4, 4), dtype=float32, numpy=
+            array([[0., 0., 1., 0.],
+                   [0., 1., 0., 0.],
+                   [0., 0., 0., 1.],
+                   [1., 0., 0., 0.]], dtype=float32)>
+
+            # Accuracy
+            >>> tf.reduce_sum((pygm.hungarian(X) * X_gt)) / tf.reduce_sum(X_gt)
+            <tf.Tensor: shape=(), dtype=float32, numpy=1.0>
+
     .. note::
         If you find this graph matching solver useful in your research, please cite:
 
@@ -723,7 +899,7 @@ def ipfp(K, n1=None, n2=None, n1max=None, n2max=None, x0=None,
     try:
         mod = importlib.import_module(f'pygmtools.{backend}_backend')
         fn = mod.ipfp
-    except ModuleNotFoundError and AttributeError:
+    except (ModuleNotFoundError, AttributeError):
         raise NotImplementedError(
             NOT_IMPLEMENTED_MSG.format(backend)
         )
