@@ -2,6 +2,16 @@ r"""
 The implementations of data loading and data processing.
 """
 
+# Copyright (c) 2022 Thinklab@SJTU
+# pygmtools is licensed under Mulan PSL v2.
+# You can use this software according to the terms and conditions of the Mulan PSL v2.
+# You may obtain a copy of Mulan PSL v2 at:
+# http://license.coscl.org.cn/MulanPSL2
+# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
+# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
+# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
+# See the Mulan PSL v2 for more details.
+
 import requests
 import os
 import zipfile
@@ -20,6 +30,7 @@ import json
 import scipy.io as sio
 import glob
 import random
+from pygmtools.utils import download
 
 
 VOC2011_KPT_NAMES = {
@@ -182,13 +193,16 @@ class PascalVOC:
 
         self.process()
 
-    def download(self, url=None, name=None):
+    def download(self, url=None, name=None, retries=5):
         r"""
         Automatically download PascalVOC dataset.
 
         :param url: str, web url of PascalVOC and PascalVOC annotation
         :param name: str, ``"PascalVOC"`` to download PascalVOC and ``"PascalVOC_anno"`` to download PascalVOC annotation
         """
+        if retries <= 0:
+            raise RuntimeError('Max Retries exceeded!')
+
         dirs = 'data/'
         if not os.path.exists(dirs):
             os.makedirs(dirs)
@@ -196,12 +210,14 @@ class PascalVOC:
         if name == "PascalVOC_anno":
             print('Downloading dataset annotation...')
             filename = "data/PascalVOC.tgz"
-            down_res = requests.get(url, stream=True)
-            file_size = int(down_res.headers.get('Content-Length', 0))
-            with tqdm.wrapattr(down_res.raw, "read", total=file_size) as content:
-                with open(filename, 'wb') as file:
-                        shutil.copyfileobj(content, file)
-            tar = tarfile.open(filename, "r")
+            download(filename=filename, url=url, to_cache=False)
+            try:
+                tar = tarfile.open(filename, "r")
+            except tarfile.ReadError as err:
+                print('Warning: Content error. Retrying...\n', err)
+                os.remove(filename)
+                return self.download(url, name, retries - 1)
+
             file_names = tar.getnames()
             print('Unzipping files...')
             sleep(0.5)
@@ -213,12 +229,14 @@ class PascalVOC:
         if name == "PascalVOC":
             print('Downloading dataset PascalVOC...')
             filename = "data/PascalVOC.tar"
-            down_res = requests.get(url, stream=True)
-            file_size = int(down_res.headers.get('Content-Length', 0))
-            with tqdm.wrapattr(down_res.raw, "read", total=file_size) as content:
-                with open(filename, 'wb') as file:
-                        shutil.copyfileobj(content, file)
-            tar = tarfile.open(filename, "r")
+            download(filename=filename, url=url, to_cache=False)
+            try:
+                tar = tarfile.open(filename, "r")
+            except tarfile.ReadError as err:
+                print('Warning: Content error. Retrying...\n', err)
+                os.remove(filename)
+                return self.download(url, name, retries - 1)
+
             file_names = tar.getnames()
             print('Unzipping files...')
             sleep(0.5)
@@ -226,6 +244,7 @@ class PascalVOC:
                 tar.extract(file_name, "data/PascalVOC/")
             tar.close()
             os.remove(filename)
+        return filename
 
     def __filter_list(self, a_xml_list):
         """
@@ -446,6 +465,7 @@ class WillowObject:
         SPLIT_OFFSET = dataset_cfg.WillowObject.SPLIT_OFFSET
         TRAIN_SAME_AS_TEST = dataset_cfg.WillowObject.TRAIN_SAME_AS_TEST
         RAND_OUTLIER = dataset_cfg.WillowObject.RAND_OUTLIER
+        URL = 'http://www.di.ens.fr/willow/research/graphlearning/WILLOW-ObjectClass_dataset.zip'
         if len(ds_dict.keys()) > 0:
             if 'CLASSES' in ds_dict.keys():
                 CLASSES = ds_dict['CLASSES']
@@ -459,11 +479,13 @@ class WillowObject:
                 TRAIN_SAME_AS_TEST = ds_dict['TRAIN_SAME_AS_TEST']
             if 'RAND_OUTLIER' in ds_dict.keys():
                 RAND_OUTLIER = ds_dict['RAND_OUTLIER']
+            if 'URL' in ds_dict.keys():
+                URL = ds_dict['URL']
 
         self.dataset_dir = 'data/WillowObject'
         if not os.path.exists(ROOT_DIR):
             assert ROOT_DIR == dataset_cfg.WillowObject.ROOT_DIR, 'you should not change ROOT_DIR unless the data have been manually downloaded'
-            self.download(url='http://www.di.ens.fr/willow/research/graphlearning/WILLOW-ObjectClass_dataset.zip')
+            self.download(url=URL)
 
         if not os.path.exists(self.dataset_dir):
             os.makedirs(self.dataset_dir)
@@ -488,29 +510,35 @@ class WillowObject:
 
         self.process()
 
-    def download(self, url=None):
+    def download(self, url=None, retries=5):
         r"""
          Automatically download WillowObject dataset.
 
          :param url: str, web url of WillowObject
          """
+        if retries <= 0:
+            raise RuntimeError('Max Retries exceeded!')
+
         dirs = 'data/'
         if not os.path.exists(dirs):
             os.makedirs(dirs)
 
         print('Downloading dataset WillowObject...')
         filename = "data/WILLOW.zip"
-        down_res = requests.get(url, stream=True)
-        file_size = 68635332
-        with tqdm.wrapattr(down_res.raw, "read", total=file_size) as content:
-            with open(filename, 'wb') as file:
-                shutil.copyfileobj(content, file)
-        fz = zipfile.ZipFile(filename, "r")
+        download(filename=filename, url=url, to_cache=False)
+        try:
+            fz = zipfile.ZipFile(filename, "r")
+        except zipfile.BadZipFile as err:
+            print('Warning: Content error. Retrying...\n', err)
+            os.remove(filename)
+            return self.download(url, retries - 1)
+
         print('Unzipping files...')
         sleep(0.5)
         for file in tqdm(fz.namelist()):
             fz.extract(file, "data/WillowObject/")
         os.remove(filename)
+        return filename
 
     def process(self):
         r"""
@@ -595,23 +623,23 @@ class WillowObject:
         if self.sets == 'train':
             for x in range(len(self.mat_list)):
                 for name in self.mat_list[x]:
-                    tmp = str(name).split('/')
+                    tmp = os.path.split(str(name))
                     objID = tmp[-1].split('.')[0]
                     train_list.append(objID)
             for x in range(len(mat_list_)):
                 for name in mat_list_[x]:
-                    tmp = str(name).split('/')
+                    tmp = os.path.split(str(name))
                     objID = tmp[-1].split('.')[0]
                     test_list.append(objID)
         else:
             for x in range(len(self.mat_list)):
                 for name in self.mat_list[x]:
-                    tmp = str(name).split('/')
+                    tmp = os.path.split(str(name))
                     objID = tmp[-1].split('.')[0]
                     test_list.append(objID)
             for x in range(len(mat_list_)):
                 for name in mat_list_[x]:
-                    tmp = str(name).split('/')
+                    tmp = os.path.split(str(name))
                     objID = tmp[-1].split('.')[0]
                     train_list.append(objID)
         str1 = json.dumps(train_list)
@@ -628,9 +656,9 @@ class WillowObject:
 
             for x in range(len(data_list)):
                 for name in data_list[x]:
-                    tmp = str(name).split('/')
+                    tmp = os.path.split(str(name))
                     objID = tmp[-1].split('.')[0]
-                    cls = tmp[3]
+                    cls = os.path.split(tmp[0])[-1]
                     annotations = self.__get_anno_dict(name, cls)
                     data_dict[objID] = annotations
 
@@ -752,7 +780,7 @@ class SPair71k:
         self.classes = list(map(lambda x: os.path.basename(x), glob.glob("%s/*" % SPair71k_image_path)))
         self.classes.sort()
         self.combine_classes = COMB_CLS
-        self.ann_files_filtered, self.ann_files_filtered_cls_dict, self.classes = self.__filter_annotations(
+        self.ann_files_filtered, self.ann_files_filtered_cls_dict, _ = self.__filter_annotations(
             self.ann_files, self.difficulty_params
         )
         self.total_size = len(self.ann_files_filtered)
@@ -760,23 +788,28 @@ class SPair71k:
 
         self.process()
 
-    def download(self, url=None):
+    def download(self, url=None, retries=5):
         r"""
          Automatically download SPair71k dataset.
 
          :param url: str, web url of SPair71k
          """
+        if retries <= 0:
+            raise RuntimeError('Max Retries exceeded!')
+
         dirs = 'data/'
         if not os.path.exists(dirs):
             os.makedirs(dirs)
         print('Downloading dataset SPair-71k...')
         filename = "data/SPair-71k.tgz"
-        down_res = requests.get(url, stream=True)
-        file_size = int(down_res.headers.get('Content-Length', 0))
-        with tqdm.wrapattr(down_res.raw, "read", total=file_size) as content:
-            with open(filename, 'wb') as file:
-                shutil.copyfileobj(content, file)
-        tar = tarfile.open(filename, "r")
+        download(filename=filename, url=url, to_cache=False)
+        try:
+            tar = tarfile.open(filename, "r")
+        except tarfile.ReadError as err:
+            print('Warning: Content error. Retrying...\n', err)
+            os.remove(filename)
+            return self.download(url, retries - 1)
+
         file_names = tar.getnames()
         print('Unzipping files...')
         sleep(0.5)
@@ -784,6 +817,7 @@ class SPair71k:
             tar.extract(file_name, "data/")
         tar.close()
         os.remove(filename)
+        return filename
 
     def process(self):
         r"""
@@ -793,7 +827,7 @@ class SPair71k:
         train_file = os.path.join(self.dataset_dir, 'train.json')
         test_file = os.path.join(self.dataset_dir, 'test.json')
         img_file = os.path.join(self.dataset_dir, 'data-' + str(self.obj_resize) + '-' + self.suffix + '.json')
-        if not (os.path.exists(train_file) and os.path.exists(test_file) and os.path.exists(img_file)):
+        if (not os.path.exists(train_file)) or (not os.path.exists(test_file)):
             train_list = []
             test_list = []
             if self.sets == 'trn':
@@ -805,18 +839,10 @@ class SPair71k:
                     pair_tuple = (id1, id2)
                     train_list.append(pair_tuple)
 
-                ann_files_ = open(os.path.join(self.SPair71k_layout_path, self.SPair71k_dataset_size + "/test.txt"),
-                                      "r").read().split("\n")
-                ann_files_ = ann_files_[: len(ann_files_) - 1]
-                ann_files_filtered_ = self.__filter_annotations(ann_files_, self.difficulty_params)[0]
-                for x in ann_files_filtered_:
-                    tmp = x.split('-')
-                    tmp2 = tmp[2].split(':')
-                    id1 = tmp[1] + '_' + tmp2[1]
-                    id2 = tmp2[0] + '_' + tmp2[1]
-                    pair_tuple = (id1, id2)
-                    test_list.append(pair_tuple)
-
+                str1 = json.dumps(train_list)
+                f1 = open(train_file, 'w')
+                f1.write(str1)
+                f1.close()
             else:
                 for x in self.ann_files_filtered:
                     tmp = x.split('-')
@@ -826,44 +852,25 @@ class SPair71k:
                     pair_tuple = (id1, id2)
                     test_list.append(pair_tuple)
 
-                ann_files_ = open(os.path.join(self.SPair71k_layout_path, self.SPair71k_dataset_size + "/trn.txt"),
-                                      "r").read().split("\n")
-                ann_files_ = ann_files_[: len(ann_files_) - 1]
-                ann_files_filtered_ = self.__filter_annotations(ann_files_, self.difficulty_params)[0]
-                for x in ann_files_filtered_:
-                    tmp = x.split('-')
-                    tmp2 = tmp[2].split(':')
-                    id1 = tmp[1] + '_' + tmp2[1]
-                    id2 = tmp2[0] + '_' + tmp2[1]
-                    pair_tuple = (id1, id2)
-                    train_list.append(pair_tuple)
-
-            str1 = json.dumps(train_list)
-            f1 = open(train_file, 'w')
-            f1.write(str1)
-            f1.close()
-            str2 = json.dumps(test_list)
-            f2 = open(test_file, 'w')
-            f2.write(str2)
-            f2.close()
+                str2 = json.dumps(test_list)
+                f2 = open(test_file, 'w')
+                f2.write(str2)
+                f2.close()
 
             data_list = []
             data_dict = dict()
             for cls_name in self.classes:
                 cls_json_list = [p for p in (self.image_annoation / cls_name).glob('*.json')]
                 ori_len = len(cls_json_list)
-                assert ori_len > 0, 'No data found for WILLOW Object Class. Is the dataset installed correctly?'
+                assert ori_len > 0, 'No data found for SPair-71k. Is the dataset installed correctly?'
                 data_list.append(cls_json_list)
 
-            list00 = []
             for x in range(len(data_list)):
                 for name in data_list[x]:
-                    tmp = str(name).split('/')
+                    tmp = os.path.split(str(name))
                     objID = tmp[-1].split('.')[0]
-                    cls = tmp[3]
+                    cls = os.path.split(tmp[0])[-1]
                     annotations = self.__get_anno_dict(name, cls)
-                    if objID in data_dict.keys():
-                        list00.append(objID)
                     ID = objID + '_' + cls
                     data_dict[ID] = annotations
 
@@ -872,7 +879,6 @@ class SPair71k:
             f3.write(data_str)
             f3.close()
 
-
     def __get_anno_dict(self, anno_file, cls):
         assert anno_file.exists(), '{} does not exist.'.format(anno_file)
 
@@ -995,6 +1001,7 @@ class IMC_PT_SparseGM:
         CLASSES = dataset_cfg.IMC_PT_SparseGM.CLASSES
         ROOT_DIR_NPZ = dataset_cfg.IMC_PT_SparseGM.ROOT_DIR_NPZ
         ROOT_DIR_IMG = dataset_cfg.IMC_PT_SparseGM.ROOT_DIR_IMG
+        URL = 'https://drive.google.com/u/0/uc?export=download&confirm=Z-AR&id=1Po9pRMWXTqKK2ABPpVmkcsOq-6K_2v-B'
         if len(ds_dict.keys()) > 0:
             if 'MAX_KPT_NUM' in ds_dict.keys():
                 MAX_KPT_NUM = ds_dict['MAX_KPT_NUM']
@@ -1004,17 +1011,20 @@ class IMC_PT_SparseGM:
                 ROOT_DIR_NPZ = ds_dict['ROOT_DIR_NPZ']
             if 'ROOT_DIR_IMG' in ds_dict.keys():
                 ROOT_DIR_IMG = ds_dict['ROOT_DIR_IMG']
+            if 'URL' in ds_dict.keys():
+                URL = ds_dict['URL']
 
         self.dataset_dir = 'data/IMC-PT-SparseGM'
         if not os.path.exists(ROOT_DIR_IMG):
             assert ROOT_DIR_IMG == dataset_cfg.IMC_PT_SparseGM.ROOT_DIR_IMG, 'you should not change ROOT_DIR_IMG or ROOT_DIR_NPZ unless the data have been manually downloaded'
             assert ROOT_DIR_NPZ == dataset_cfg.IMC_PT_SparseGM.ROOT_DIR_NPZ, 'you should not change ROOT_DIR_IMG or ROOT_DIR_NPZ unless the data have been manually downloaded'
-            self.download(url='https://drive.google.com/u/0/uc?export=download&confirm=Z-AR&id=1Po9pRMWXTqKK2ABPpVmkcsOq-6K_2v-B')
+            self.download(url=URL)
 
         if not os.path.exists(self.dataset_dir):
             os.makedirs(self.dataset_dir)
         self.sets = sets
         self.classes = CLASSES[sets]
+        self.class_dict = CLASSES
         self.max_kpt_num = MAX_KPT_NUM
         self.suffix = 'imcpt-' + str(MAX_KPT_NUM)
 
@@ -1027,23 +1037,28 @@ class IMC_PT_SparseGM:
 
         self.process()
 
-    def download(self, url=None):
+    def download(self, url=None, retries=15):
         r"""
          Automatically download IMC_PT_SparseGM dataset.
 
          :param url: str, web url of IMC_PT_SparseGM
          """
+        if retries <= 0:
+            raise RuntimeError('Max Retries exceeded!')
+
         dirs = 'data/'
         if not os.path.exists(dirs):
             os.makedirs(dirs)
         print('Downloading dataset IMC-PT-SparseGM...')
         filename = 'data/IMC-PT-SparseGM.tar.gz'
-        down_res = requests.get(url, stream=True)
-        file_size = int(down_res.headers.get('Content-Length', 0))
-        with tqdm.wrapattr(down_res.raw, "read", total=file_size) as content:
-            with open(filename, 'wb') as file:
-                shutil.copyfileobj(content, file)
-        tar = tarfile.open(filename, "r")
+        download(filename=filename, url=url, to_cache=False)
+        try:
+            tar = tarfile.open(filename, "r")
+        except tarfile.ReadError as err:
+            print('Warning: Content error. Retrying...\n', err)
+            os.remove(filename)
+            return self.download(url, retries - 1)
+
         file_names = tar.getnames()
         print('Unzipping files...')
         sleep(0.5)
@@ -1051,6 +1066,7 @@ class IMC_PT_SparseGM:
             tar.extract(file_name, "data/")
         tar.close()
         os.remove(filename)
+        return filename
 
     def process(self):
         r"""
@@ -1072,9 +1088,9 @@ class IMC_PT_SparseGM:
 
         if not os.path.exists(img_file):
             total_cls = []
-            for cls in dataset_cfg.IMC_PT_SparseGM.CLASSES['train']:
+            for cls in self.class_dict['train']:
                 total_cls.append(cls)
-            for cls in dataset_cfg.IMC_PT_SparseGM.CLASSES['test']:
+            for cls in self.class_dict['test']:
                 total_cls.append(cls)
 
             total_img_lists = [np.load(self.root_path_npz / cls / 'img_info.npz')['img_name'].tolist()
@@ -1147,20 +1163,19 @@ class CUB2011:
 
     :param sets: str, problem set, ``'train'`` for training set and ``'test'`` for testing set
     :param obj_resize: tuple, resized image size
-    :param ds_dict: settings of dataset, containing at most 2 params(keys) for CUB2011:
+    :param ds_dict: settings of dataset, containing at most 1 params(key) for CUB2011:
 
         * **ROOT_DIR**: str, directory of data
-
-        * **CLS_SPLIT**: str, ``'ori'`` (original split), ``'sup'`` (super class) or ``'all'`` (all birds as one class)
     """
     def __init__(self, sets, obj_resize, **ds_dict):
         CLS_SPLIT = dataset_cfg.CUB2011.CLASS_SPLIT
         ROOT_DIR = dataset_cfg.CUB2011.ROOT_DIR
+        URL = 'https://drive.google.com/u/0/uc?export=download&confirm=B8eu&id=1hbzc_P1FuxMkcabkgn9ZKinBwW683j45'
         if len(ds_dict.keys()) > 0:
-            if 'CLS_SPLIT' in ds_dict.keys():
-                CLS_SPLIT = ds_dict['CLS_SPLIT']
             if 'ROOT_DIR' in ds_dict.keys():
                 ROOT_DIR = ds_dict['ROOT_DIR']
+            if 'URL' in ds_dict.keys():
+                URL = ds_dict['URL']
 
         self.set_data = {'train': [], 'test': []}
         self.classes = []
@@ -1168,14 +1183,14 @@ class CUB2011:
         self._set_pairs = {}
         self._set_mask = {}
         self.cls_split = CLS_SPLIT
-        self.suffix = 'cub2011-' + CLS_SPLIT
+        self.suffix = 'cub2011'
 
         self.rootpath = ROOT_DIR
 
         self.dataset_dir = 'data/CUB_200_2011'
         if not os.path.exists(ROOT_DIR):
             assert ROOT_DIR == dataset_cfg.CUB2011.ROOT_DIR, 'you should not change ROOT_DIR unless the data have been manually downloaded'
-            self.download(url='https://drive.google.com/u/0/uc?export=download&confirm=B8eu&id=1hbzc_P1FuxMkcabkgn9ZKinBwW683j45')
+            self.download(url=URL)
 
         if not os.path.exists(self.dataset_dir):
             os.makedirs(self.dataset_dir)
@@ -1210,61 +1225,33 @@ class CUB2011:
                         test_set.append(img_idx)
                 self.set_data['train'].append(train_set)
                 self.set_data['test'].append(test_set)
-        elif self.cls_split == 'sup':
-            super_classes = [v.split('_')[-1] for v in classes.values()]
-            self.classes = list(set(super_classes))
-            for cls in self.classes:
-                self.set_data['train'].append([])
-                self.set_data['test'].append([])
-            for class_idx in sorted(classes):
-                supcls_idx = self.classes.index(classes[class_idx].split('_')[-1])
-                train_set = []
-                test_set = []
-                for img_idx in class2img[class_idx]:
-                    if train_split[img_idx] == '1':
-                        train_set.append(img_idx)
-                    else:
-                        test_set.append(img_idx)
-                self.set_data['train'][supcls_idx] += train_set
-                self.set_data['test'][supcls_idx] += test_set
-        elif self.cls_split == 'all':
-            self.classes.append('cub2011')
-            self.set_data['train'].append([])
-            self.set_data['test'].append([])
-            for class_idx in sorted(classes):
-                train_set = []
-                test_set = []
-                for img_idx in class2img[class_idx]:
-                    if train_split[img_idx] == '1':
-                        train_set.append(img_idx)
-                    else:
-                        test_set.append(img_idx)
-                self.set_data['train'][0] += train_set
-                self.set_data['test'][0] += test_set
-        else:
-            raise ValueError('Unknown CUB2011.CLASS_SPLIT {}'.format(self.cls_split))
         self.sets = sets
         self.obj_resize = obj_resize
 
         self.process()
 
-    def download(self, url=None):
+    def download(self, url=None, retries=50):
         r"""
          Automatically download CUB2011 dataset.
 
          :param url: str, web url of CUB2011
          """
+        if retries <= 0:
+            raise RuntimeError('Max Retries exceeded!')
+
         dirs = 'data/'
         if not os.path.exists(dirs):
             os.makedirs(dirs)
         print('Downloading dataset CUB2011...')
         filename = 'data/CUB_200_2011.tgz'
-        down_res = requests.get(url, stream=True)
-        file_size = int(down_res.headers.get('Content-Length', 0))
-        with tqdm.wrapattr(down_res.raw, "read", total=file_size) as content:
-            with open(filename, 'wb') as file:
-                shutil.copyfileobj(content, file)
-        tar = tarfile.open(filename, "r")
+        download(filename=filename, url=url, to_cache=False)
+        try:
+            tar = tarfile.open(filename, "r")
+        except tarfile.ReadError as err:
+            print('Warning: Content error. Retrying...\n', err)
+            os.remove(filename)
+            return self.download(url, retries - 1)
+
         file_names = tar.getnames()
         print('Unzipping files...')
         sleep(0.5)
@@ -1272,6 +1259,7 @@ class CUB2011:
             tar.extract(file_name, "data/")
         tar.close()
         os.remove(filename)
+        return filename
 
     def process(self):
         r"""
