@@ -1,3 +1,13 @@
+# Copyright (c) 2022 Thinklab@SJTU
+# pygmtools is licensed under Mulan PSL v2.
+# You can use this software according to the terms and conditions of the Mulan PSL v2.
+# You may obtain a copy of Mulan PSL v2 at:
+# http://license.coscl.org.cn/MulanPSL2
+# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
+# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
+# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
+# See the Mulan PSL v2 for more details.
+
 from easydict import EasyDict as edict
 
 __C = edict()
@@ -24,8 +34,8 @@ __C.WillowObject.RAND_OUTLIER = 0
 
 # CUB2011 dataset
 __C.CUB2011 = edict()
-__C.CUB2011.ROOT_DIR = 'data/CUB_200_2011/CUB_200_2011'
-__C.CUB2011.CLASS_SPLIT = 'ori' # choose from 'ori' (original split), 'sup' (super class) or 'all' (all birds as one class)
+__C.CUB2011.ROOT_DIR = 'data/CUB_200_2011'
+__C.CUB2011.CLASS_SPLIT = 'ori' # choose from 'ori' (original split), 'sup' (super class) or 'all' (all birds as one class), only support 'ori'
 
 # SWPair-71 Dataset
 __C.SPair = edict()
