@@ -1,10 +1,20 @@
+# Copyright (c) 2022 Thinklab@SJTU
+# pygmtools is licensed under Mulan PSL v2.
+# You can use this software according to the terms and conditions of the Mulan PSL v2.
+# You may obtain a copy of Mulan PSL v2 at:
+# http://license.coscl.org.cn/MulanPSL2
+# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
+# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
+# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
+# See the Mulan PSL v2 for more details.
+
 import itertools
 import functools
 import torch
 import numpy as np
 from multiprocessing import Pool
 from torch import Tensor
-
+import os
 import pygmtools.utils
 
 
@@ -421,7 +431,7 @@ def cao_fast_solver(K, X, num_graph, num_node, max_iter, lambda_init, lambda_ste
 
         score_combo, idx = torch.max(score_combo, dim=-1)
 
-        assert torch.all(score_combo >= score_ori), torch.min(score_combo - score_ori)
+        assert torch.all(score_combo + 1e-4 >= score_ori), torch.min(score_combo - score_ori)
         X_upt = X_combo[mask1, mask2, idx, :, :]
         X = X_upt * X_mask + X_upt.transpose(0, 1).transpose(2, 3) * X_mask.transpose(0, 1) + X * (1 - X_mask - X_mask.transpose(0, 1))
         assert torch.all(X.transpose(0, 1).transpose(2, 3) == X)
@@ -1096,11 +1106,11 @@ class NGM_Net(torch.nn.Module):
             if i == 0:
                 gnn_layer = NGMConvLayer(1, 1,
                                          gnn_channels[i] + sk_emb, gnn_channels[i],
-                                         sk_channel=sk_emb, edge_emb=False)
+                                         sk_channel=sk_emb)
             else:
                 gnn_layer = NGMConvLayer(gnn_channels[i - 1] + sk_emb, gnn_channels[i - 1],
                                          gnn_channels[i] + sk_emb, gnn_channels[i],
-                                         sk_channel=sk_emb, edge_emb=False)
+                                         sk_channel=sk_emb)
             self.add_module('gnn_layer_{}'.format(i), gnn_layer)
         self.classifier = nn.Linear(gnn_channels[-1] + sk_emb, 1)
 
@@ -1146,7 +1156,10 @@ def ngm(K, n1, n2, n1max, n2max, x0, gnn_channels, sk_emb, sk_max_iter, sk_tau,
         if pretrain:
             if pretrain in ngm_pretrain_path:
                 url, md5 = ngm_pretrain_path[pretrain]
-                filename = pygmtools.utils.download(f'ngm_{pretrain}_pytorch.pt', url, md5)
+                try:
+                    filename = pygmtools.utils.download(f'ngm_{pretrain}_pytorch.pt', url, md5)
+                except:
+                    filename = os.path.dirname(__file__) + f'/temp/ngm_{pretrain}_pytorch.pt'
                 _load_model(network, filename, device)
             else:
                 raise ValueError(f'Unknown pretrain tag. Available tags: {ngm_pretrain_path.keys()}')
@@ -1361,13 +1374,14 @@ def _aff_mat_from_node_edge_aff(node_aff: Tensor, edge_aff: Tensor, connectivity
     return torch.stack(ks, dim=0)
 
 
-def _check_data_type(input: Tensor, var_name=None):
+def _check_data_type(input: Tensor, var_name, raise_err):
     """
     Pytorch implementation of _check_data_type
     """
-    if type(input) is not Tensor:
-        raise ValueError(f'Expected Pytorch Tensor{f" for variable {var_name}" if var_name is not None else ""}, '
+    if raise_err and type(input) is not Tensor:
+        raise ValueError(f'Expected PyTorch Tensor{f" for variable {var_name}" if var_name is not None else ""}, '
                          f'but got {type(input)}. Perhaps the wrong backend?')
+    return type(input) is Tensor
 
 
 def _check_shape(input, dim_num):
