@@ -2,6 +2,16 @@
 Utility functions: problem formulating, data processing, and beyond.
 """
 
+# Copyright (c) 2022 Thinklab@SJTU
+# pygmtools is licensed under Mulan PSL v2.
+# You can use this software according to the terms and conditions of the Mulan PSL v2.
+# You may obtain a copy of Mulan PSL v2 at:
+# http://license.coscl.org.cn/MulanPSL2
+# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
+# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
+# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
+# See the Mulan PSL v2 for more details.
+
 import time
 import functools
 import importlib
@@ -13,12 +23,13 @@ import hashlib
 import shutil
 from tqdm.auto import tqdm
 import inspect
-
+import wget
+import numpy as np
 import pygmtools
 
 NOT_IMPLEMENTED_MSG = \
     'The backend function for {} is not implemented. ' \
-    'If you are a user, please use other backends as workarounds.' \
+    'If you are a user, please check the spelling, and use other backends as workarounds. ' \
     'If you are a developer, it will be truly appreciated if you could develop and share your' \
     ' implementation with the community! See our Github: https://github.com/Thinklab-SJTU/pygmtools'
 
@@ -173,6 +184,8 @@ def build_aff_mat(node_feat1, edge_feat1, connectivity1, node_feat2, edge_feat2,
         if all([_check_shape(_, 2, backend) for _ in (node_feat1, node_feat2)]):
             non_batched_input = True
             node_feat1, node_feat2 = [_unsqueeze(_, 0, backend) for _ in (node_feat1, node_feat2)]
+            if type(n1) is int: n1 = from_numpy(np.array([n1]), backend=backend)
+            if type(n2) is int: n2 = from_numpy(np.array([n2]), backend=backend)
         elif all([_check_shape(_, 3, backend) for _ in (node_feat1, node_feat2)]):
             pass
         else:
@@ -191,6 +204,8 @@ def build_aff_mat(node_feat1, edge_feat1, connectivity1, node_feat2, edge_feat2,
             non_batched_input = True
             edge_feat1, edge_feat2, connectivity1, connectivity2 = \
                 [_unsqueeze(_, 0, backend) for _ in (edge_feat1, edge_feat2, connectivity1, connectivity2)]
+            if type(ne1) is int: ne1 = from_numpy(np.array([ne1]), backend=backend)
+            if type(ne2) is int: ne2 = from_numpy(np.array([ne2]), backend=backend)
         elif all([_check_shape(_, 3, backend) for _ in (edge_feat1, edge_feat2, connectivity1, connectivity2)]):
             pass
         else:
@@ -215,7 +230,8 @@ def build_aff_mat(node_feat1, edge_feat1, connectivity1, node_feat2, edge_feat2,
     node_aff = node_aff_fn(node_feat1, node_feat2) if node_feat1 is not None else None
     edge_aff = edge_aff_fn(edge_feat1, edge_feat2) if edge_feat1 is not None else None
 
-    result = _aff_mat_from_node_edge_aff(node_aff, edge_aff, connectivity1, connectivity2, n1, n2, ne1, ne2, backend=backend)
+    result = _aff_mat_from_node_edge_aff(node_aff, edge_aff, connectivity1, connectivity2, n1, n2, ne1, ne2,
+                                         backend=backend)
     if non_batched_input:
         return _squeeze(result, 0, backend)
     else:
@@ -243,7 +259,7 @@ def inner_prod_aff_fn(feat1, feat2, backend=None):
     try:
         mod = importlib.import_module(f'pygmtools.{backend}_backend')
         fn = mod.inner_prod_aff_fn
-    except ModuleNotFoundError and AttributeError:
+    except (ModuleNotFoundError, AttributeError):
         raise NotImplementedError(
             NOT_IMPLEMENTED_MSG.format(backend)
         )
@@ -272,7 +288,7 @@ def gaussian_aff_fn(feat1, feat2, sigma=1., backend=None):
     try:
         mod = importlib.import_module(f'pygmtools.{backend}_backend')
         fn = mod.gaussian_aff_fn
-    except ModuleNotFoundError and AttributeError:
+    except (ModuleNotFoundError, AttributeError):
         raise NotImplementedError(
             NOT_IMPLEMENTED_MSG.format(backend)
         )
@@ -493,7 +509,7 @@ def dense_to_sparse(dense_adj, backend=None):
     try:
         mod = importlib.import_module(f'pygmtools.{backend}_backend')
         fn = mod.dense_to_sparse
-    except ModuleNotFoundError and AttributeError:
+    except (ModuleNotFoundError, AttributeError):
         raise NotImplementedError(
             NOT_IMPLEMENTED_MSG.format(backend)
         )
@@ -566,7 +582,7 @@ def compute_affinity_score(X, K, backend=None):
     try:
         mod = importlib.import_module(f'pygmtools.{backend}_backend')
         fn = mod.compute_affinity_score
-    except ModuleNotFoundError and AttributeError:
+    except (ModuleNotFoundError, AttributeError):
         raise NotImplementedError(
             NOT_IMPLEMENTED_MSG.format(backend)
         )
@@ -598,7 +614,7 @@ def to_numpy(input, backend=None):
         try:
             mod = importlib.import_module(f'pygmtools.{backend}_backend')
             fn = mod.to_numpy
-        except ModuleNotFoundError and AttributeError:
+        except (ModuleNotFoundError, AttributeError):
             raise NotImplementedError(
                 NOT_IMPLEMENTED_MSG.format(backend)
             )
@@ -626,7 +642,7 @@ def from_numpy(input, device=None, backend=None):
         try:
             mod = importlib.import_module(f'pygmtools.{backend}_backend')
             fn = mod.from_numpy
-        except ModuleNotFoundError and AttributeError:
+        except (ModuleNotFoundError, AttributeError):
             raise NotImplementedError(
                 NOT_IMPLEMENTED_MSG.format(backend)
             )
@@ -654,7 +670,7 @@ def generate_isomorphic_graphs(node_num, graph_num=2, node_feat_dim=0, backend=N
     try:
         mod = importlib.import_module(f'pygmtools.{backend}_backend')
         fn = mod.generate_isomorphic_graphs
-    except ModuleNotFoundError and AttributeError:
+    except (ModuleNotFoundError, AttributeError):
         raise NotImplementedError(
             NOT_IMPLEMENTED_MSG.format(backend)
         )
@@ -888,7 +904,7 @@ def permutation_loss(pred_dsmat, gt_perm, n1=None, n2=None, backend=None):
     try:
         mod = importlib.import_module(f'pygmtools.{backend}_backend')
         fn = mod.permutation_loss
-    except ModuleNotFoundError and AttributeError:
+    except (ModuleNotFoundError, AttributeError):
         raise NotImplementedError(
             NOT_IMPLEMENTED_MSG.format(backend)
         )
@@ -927,28 +943,49 @@ def _aff_mat_from_node_edge_aff(node_aff, edge_aff, connectivity1, connectivity2
     try:
         mod = importlib.import_module(f'pygmtools.{backend}_backend')
         fn = mod._aff_mat_from_node_edge_aff
-    except ModuleNotFoundError and AttributeError:
+    except (ModuleNotFoundError, AttributeError):
         raise NotImplementedError(
             NOT_IMPLEMENTED_MSG.format(backend)
         )
     return fn(*args)
 
 
-def _check_data_type(input, var_name=None, backend=None):
+def _check_data_type(input, *args):
     r"""
     Check whether the input data meets the backend. If not met, it will raise an ValueError
+    Three overloads of this function:
+    _check_data_type(input, backend)
+    _check_data_type(input, var_name, backend)
+    _check_data_type(input, var_name, raise_err, backend)
 
     :param input: input data (must be Tensor/ndarray)
     :param var_name: name of the variable
-    :return: None
+    :param raise_err: raise an error if input data not true
+    :return: True or False
     """
+    if len(args) == 3:
+        var_name, raise_err, backend = args
+    elif len(args) == 2:
+        var_name, backend = args
+        raise_err = True
+    elif len(args) == 1:
+        backend = args[0]
+        var_name = None
+        raise_err = True
+    elif len(args) == 0:
+        backend = None
+        var_name = None
+        raise_err = True
+    else:
+        raise RuntimeError(f'Unknown arguments: {args}')
+
     if backend is None:
         backend = pygmtools.BACKEND
-    args = (input, var_name)
+    args = (input, var_name, raise_err)
     try:
         mod = importlib.import_module(f'pygmtools.{backend}_backend')
         fn = mod._check_data_type
-    except ModuleNotFoundError and AttributeError:
+    except (ModuleNotFoundError, AttributeError):
         raise NotImplementedError(
             NOT_IMPLEMENTED_MSG.format(backend)
         )
@@ -969,7 +1006,7 @@ def _check_shape(input, num_dim, backend=None):
     try:
         mod = importlib.import_module(f'pygmtools.{backend}_backend')
         fn = mod._check_shape
-    except ModuleNotFoundError and AttributeError:
+    except (ModuleNotFoundError, AttributeError):
         raise NotImplementedError(
             NOT_IMPLEMENTED_MSG.format(backend)
         )
@@ -989,7 +1026,7 @@ def _get_shape(input, backend=None):
     try:
         mod = importlib.import_module(f'pygmtools.{backend}_backend')
         fn = mod._get_shape
-    except ModuleNotFoundError and AttributeError:
+    except (ModuleNotFoundError, AttributeError):
         raise NotImplementedError(
             NOT_IMPLEMENTED_MSG.format(backend)
         )
@@ -1010,7 +1047,7 @@ def _squeeze(input, dim, backend=None):
     try:
         mod = importlib.import_module(f'pygmtools.{backend}_backend')
         fn = mod._squeeze
-    except ModuleNotFoundError and AttributeError:
+    except (ModuleNotFoundError, AttributeError):
         raise NotImplementedError(
             NOT_IMPLEMENTED_MSG.format(backend)
         )
@@ -1031,7 +1068,7 @@ def _unsqueeze(input, dim, backend=None):
     try:
         mod = importlib.import_module(f'pygmtools.{backend}_backend')
         fn = mod._unsqueeze
-    except ModuleNotFoundError and AttributeError:
+    except (ModuleNotFoundError, AttributeError):
         raise NotImplementedError(
             NOT_IMPLEMENTED_MSG.format(backend)
         )
@@ -1053,7 +1090,7 @@ def _transpose(input, dim1, dim2, backend=None):
     try:
         mod = importlib.import_module(f'pygmtools.{backend}_backend')
         fn = mod._transpose
-    except ModuleNotFoundError and AttributeError:
+    except (ModuleNotFoundError, AttributeError):
         raise NotImplementedError(
             NOT_IMPLEMENTED_MSG.format(backend)
         )
@@ -1074,18 +1111,16 @@ def _mm(input1, input2, backend=None):
     try:
         mod = importlib.import_module(f'pygmtools.{backend}_backend')
         fn = mod._mm
-    except ModuleNotFoundError and AttributeError:
+    except (ModuleNotFoundError, AttributeError):
         raise NotImplementedError(
             NOT_IMPLEMENTED_MSG.format(backend)
         )
     return fn(*args)
 
-
-def download(filename, url, md5=None, retries=5):
+def download(filename, url, md5=None, retries=10, to_cache=True):
     r"""
     Check if content exits. If not, download the content to ``<user cache path>/pygmtools/<filename>``. ``<user cache path>``
     depends on your system. For example, on Debian, it should be ``$HOME/.cache``.
-
     :param filename: the destination file name
     :param url: the url
     :param md5: (optional) the md5sum to verify the content. It should match the result of ``md5sum file`` on Linux.
@@ -1095,32 +1130,45 @@ def download(filename, url, md5=None, retries=5):
     if retries <= 0:
         raise RuntimeError('Max Retries exceeded!')
 
-    dirs = user_cache_dir("pygmtools")
-    if not os.path.exists(dirs):
-        os.makedirs(dirs)
-    filename = os.path.join(dirs, filename)
+    if to_cache:
+        dirs = user_cache_dir("pygmtools")
+        if not os.path.exists(dirs):
+            os.makedirs(dirs)
+        filename = os.path.join(dirs, filename)
     if not os.path.exists(filename):
-        print(f'Downloading to {filename}...')
-        down_res = requests.get(url, stream=True)
-        file_size = int(down_res.headers.get('Content-Length', 0))
-        with tqdm.wrapattr(down_res.raw, "read", total=file_size) as content:
-            with open(filename, 'wb') as file:
-                shutil.copyfileobj(content, file)
-
+        print(f'\nDownloading to {filename}...')
+        if retries % 2 == 1:
+            try:
+                down_res = requests.get(url, stream=True)
+                file_size = int(down_res.headers.get('Content-Length', 0))
+                with tqdm.wrapattr(down_res.raw, "read", total=file_size) as content:
+                    with open(filename, 'wb') as file:
+                        shutil.copyfileobj(content, file)
+            except requests.exceptions.ConnectionError as err:
+                print('Warning: Network error. Retrying...\n', err)
+                return download(filename, url, md5, retries - 1)
+        else:
+            try:
+                wget.download(url,out=filename)
+            except:
+                return download(filename, url, md5, retries - 1)
     if md5 is not None:
-        hash_md5 = hashlib.md5()
-        chunk = 8192
-        with open(filename, 'rb') as file_to_check:
-            while True:
-                buffer = file_to_check.read(chunk)
-                if not buffer:
-                    break
-                hash_md5.update(buffer)
-            md5_returned = hash_md5.hexdigest()
+        md5_returned = _get_md5(filename)
         if md5 != md5_returned:
             print('Warning: MD5 check failed for the downloaded content. Retrying...')
             os.remove(filename)
             time.sleep(1)
             return download(filename, url, md5, retries - 1)
-
     return filename
+
+def _get_md5(filename):
+    hash_md5 = hashlib.md5()
+    chunk = 8192
+    with open(filename, 'rb') as file_to_check:
+        while True:
+            buffer = file_to_check.read(chunk)
+            if not buffer:
+                break
+            hash_md5.update(buffer)
+        md5_returned = hash_md5.hexdigest()
+        return md5_returned
